<*+M2EXTENSIONS *>
<*+NOPTRALIAS*>
<*+DOREORDER *>
<*CPU="PENTIUM"*>

IMPLEMENTATION MODULE aprspos; (* get aprs position by OE5DXL *)

FROM SYSTEM IMPORT CAST, SHIFT, ADR, CARD8, FILL;

FROM osi IMPORT WrStrLn, WrInt;

<* IF DEFINED(MYMATH) THEN *>
FROM osi IMPORT power, arccos, cos, sin, arctan, tan, sqrt;
<* ELSE *>
  <* IF TARGET_FAMILY="UNIX" THEN *>

FROM math IMPORT acos,cos,sin,atan,tan,pow,sqrt;
CONST arccos=acos;
      arctan=atan;
      power=pow;
  <* ELSIF TARGET_FAMILY="WIN32" THEN *>
FROM RealMath IMPORT arccos,cos,sin,arctan,tan,power,sqrt;
  <* END *>

<* END *>



PROCEDURE myround(x:REAL):INTEGER;
BEGIN
  IF x>=0.0 THEN x:=x+0.5 ELSE x:=x-0.5 END;
  RETURN VAL(INTEGER, x)
END myround;


PROCEDURE rad(w:REAL):REAL;
BEGIN RETURN w*(PI2/360.0) END rad;


PROCEDURE posvalid(pos-:POSITION):BOOLEAN;
BEGIN RETURN (pos.lat<>0.0) OR (pos.long<>0.0) END posvalid;

(*
PROCEDURE distance(home,dist:POSITION):REAL;
VAR h:LONGREAL;
BEGIN
  h:=ABS(dist.long-home.long);
  IF (h=0.0) & (dist.lat=home.lat) THEN RETURN 0.0
  ELSE
    IF h>PI2 THEN h:=PI2-h END;
    h:=LongMath.sin(dist.lat)*LongMath.sin(home.lat) + LongMath.cos(dist.lat)*LongMath.cos(home.lat)*LongMath.cos(h);
    IF ABS(h)>=1.0 THEN RETURN 0.0 END;
    RETURN 6370.0*LongMath.arccos(h)
  END;
END distance;
*)

PROCEDURE distance(home, dist:POSITION):REAL;
VAR x, y:REAL;
BEGIN
  x:=ABS(dist.long-home.long);
  y:=ABS(dist.lat -home.lat);
  IF (x=0.0) & (y=0.0) THEN RETURN 0.0
  ELSIF x+y<0.04 THEN                      (* near *)
    x:=x*cos((home.lat+dist.lat)*0.5);
    RETURN EARTH*sqrt(x*x+y*y)
  ELSE                                     (* far *)
    IF x>PI2 THEN x:=PI2-x END;
    x:=sin(dist.lat)*sin(home.lat) + cos(dist.lat)*cos(home.lat)*cos(x);
    IF ABS(x)>=1.0 THEN RETURN 0.0 END;
    RETURN EARTH*arccos(x)
  END;
END distance;


PROCEDURE azimuth(home,dist:POSITION):CARDINAL;
VAR h:INTEGER;
    ldiff:REAL;
BEGIN
  ldiff:=dist.long-home.long;
  IF (ldiff=0.0) OR (cos(home.lat)=0.0) OR (cos(dist.lat)=0.0) THEN
    IF home.lat<=dist.lat THEN RETURN 0 ELSE RETURN 180 END;
  ELSE
    h:=myround(360.0/PI2*arctan(cos(home.lat)*((tan(home.lat)*cos(ldiff)-tan(dist.lat))/sin(ldiff))));
    IF ldiff<0.0 THEN RETURN h+270 ELSE RETURN h+90 END;
  END;
END azimuth;



PROCEDURE GetPos(VAR pos:POSITION; VAR speed, course:CARDINAL; VAR altitude:INTEGER; 
                 VAR symb,symbt:CHAR; buf-:ARRAY OF CHAR; micedest, payload:CARDINAL;
                 VAR compos:CARDINAL; VAR postyp:CHAR);

  PROCEDURE dig(VAR s:REAL; c:CHAR; mul:REAL):BOOLEAN;
  BEGIN
    IF c=" " THEN RETURN TRUE END;                     (* ambiguity *)
    IF (c<"0") OR (c>"9") THEN RETURN FALSE END;
    s:=s + FLOAT(ORD(c)-ORD("0"))*mul;
    RETURN TRUE
  END dig;

  PROCEDURE r91(VAR s:REAL; c:CHAR; mul:REAL):BOOLEAN;
  BEGIN
    IF (c<CHR(33)) OR (c>CHR(33+91)) THEN RETURN FALSE END;
    s:=s + FLOAT(ORD(c)-33)*mul;
    RETURN TRUE
  END r91;

  PROCEDURE mic(VAR s:REAL; c:CHAR; mul:REAL):BOOLEAN;
  VAR n:CARDINAL;
  BEGIN
    IF  (c>="0") & (c<="9") THEN n:=ORD(c)-ORD("0")
    ELSIF (c>="A") & (c<="J") THEN n:=ORD(c)-ORD("A")
    ELSIF (c="K") OR (c="L") OR (c="Z") THEN n:=0
    ELSIF (c>="P") & (c<="Y") THEN n:=ORD(c)-ORD("P")
    ELSE RETURN FALSE END;
    s:=s + FLOAT(n)*mul;
    RETURN TRUE
  END mic;

  PROCEDURE bit5(c:CHAR):BOOLEAN;
  BEGIN RETURN ODD(ORD(c) DIV 32) OR (c="L") END bit5;       (* L is wildcard *)

  PROCEDURE komma(VAR p:CARDINAL);
  BEGIN WHILE (p<HIGH(buf)) & (buf[p]<>0C) & (buf[p]<>",") DO INC(p) END; INC(p) END komma;

  PROCEDURE ns(c:CHAR):BOOLEAN;
  BEGIN c:=CAP(c); RETURN (c="N") OR (c="S") END ns;

  PROCEDURE ew(c:CHAR):BOOLEAN;
  BEGIN c:=CAP(c); RETURN (c="E") OR (c="W") END ew;

  PROCEDURE dig3(VAR s, p:CARDINAL):BOOLEAN;
  VAR i:CARDINAL;
      c:CHAR;
  BEGIN
    s:=0; 
    IF (buf[p]=" ") & (buf[p+1]=" ") & (buf[p+2]=" ") 
    OR (buf[p]=".") & (buf[p+1]=".") & (buf[p+2]=".") THEN
      INC(p, 3);
      RETURN TRUE 
    END;
    FOR i:=0 TO 2 DO
      c:=buf[p]; INC(p); 
      IF (c<"0") OR (c>"9") THEN RETURN FALSE END;
      s:=s*10 + ORD(c)-ORD("0");
    END;
    RETURN TRUE
  END dig3;

  PROCEDURE dig6(VAR s:REAL; p:CARDINAL):BOOLEAN;
  VAR i:CARDINAL;
      c:CHAR;
      sig:BOOLEAN;
  BEGIN
    s:=0.0;
    IF buf[p]="-" THEN sig:=TRUE; i:=1 ELSE sig:=FALSE; i:=0 END;
    REPEAT
      c:=buf[p+i];
      IF (c<"0") OR (c>"9") THEN RETURN FALSE END;
      s:=s*10.0 + FLOAT(ORD(c)-ORD("0"));
      INC(i);
    UNTIL i>=6;
    IF sig THEN s:=-s END;
    RETURN TRUE
  END dig6;

  PROCEDURE Mapsym(c:CHAR):CHAR;
  BEGIN
    IF (c>="a") & (c<="j") THEN c:=CHR(ORD(c)-(ORD("a")-ORD("0"))) END; 
    RETURN c
  END Mapsym;


CONST DAO10=0.001*RAD/60.0;                   (* additional digit in lat/log *)
      DAO91=0.00011*RAD/60.0;                 (* additional 0..90 * 1.1 digits *)
      RND=0.005/60.0*RAD;

VAR i,len,n, nc, na:CARDINAL;
    c, gpst:CHAR;
    ok, nornd:BOOLEAN;
    sc,scl:REAL;
BEGIN
  len:=payload;
  nornd:=FALSE;
(*
IO.WrStr(" index="); IO.WrCard(len, 1); IO.WrLn;
*)
  WHILE (len<HIGH(buf)) & (buf[len]>CHR(13)) DO INC(len) END;
(*
  speed:=0;
  course:=0;
  symb:=0C;
  symbt:=0C;
*)
  compos:=0;
  c:=buf[payload];
  IF (payload+8<len) & (micedest>0) 
  & ((c=CHR(01CH)) OR (c=CHR(01DH)) OR (c=CHR(027H)) OR (c=CHR(060H))) THEN (* mic-e *)
    pos.lat:=0.0;
    i:=micedest;
    ok:=TRUE;
            IF NOT mic(pos.lat, buf[i], 10.0*RAD) THEN ok:=FALSE END;
    INC(i); IF NOT mic(pos.lat, buf[i], 1.0*RAD) THEN ok:=FALSE END;
    INC(i); IF NOT mic(pos.lat, buf[i], 10.0/60.0*RAD) THEN ok:=FALSE END;
    INC(i); IF NOT mic(pos.lat, buf[i], 1.0/60.0*RAD) THEN ok:=FALSE END;
    INC(i); IF NOT mic(pos.lat, buf[i], 0.1/60.0*RAD) THEN ok:=FALSE END;
    INC(i); IF NOT mic(pos.lat, buf[i], 0.01/60.0*RAD) THEN ok:=FALSE END;
--    IF ok & (pos.lat<>0.0) THEN pos.lat:=pos.lat+0.005/60.0*RAD END;    (* rounding *) 
    IF bit5(buf[micedest+3]) THEN pos.lat:=-pos.lat END;   (* south *)

    n:=ORD(buf[payload+1]);                                    (* long degrees *)
    IF (n>=28) & (n<=127) THEN
      DEC(n, 28);
      IF NOT bit5(buf[micedest+4]) THEN INC(n, 100) END; 
      IF (n>=180) & (n<=189) THEN DEC(n, 80)
      ELSIF (n>=190) & (n<=199) THEN DEC(n, 190) END;
    ELSE ok:=FALSE END;
    pos.long:=FLOAT(n)*RAD;

    n:=ORD(buf[payload+2]);                                    (* long min *)
    IF (n>=28) & (n<=127) THEN
      DEC(n, 28);
      IF n>=60 THEN DEC(n, 60) END;
    ELSE ok:=FALSE END;
    pos.long:=pos.long + FLOAT(n)*(RAD/60.0);


    n:=ORD(buf[payload+3]);                                    (* long min/100 *)
    IF (n>=28) & (n<=127) THEN pos.long:=pos.long + FLOAT(n-28)*(RAD/6000.0);
    ELSE ok:=FALSE END;
--    IF ok & (pos.long<>0.0) THEN pos.long:=pos.long+0.005/60.0*RAD END;    (* rounding *)
    

    IF NOT bit5(buf[micedest+5]) THEN pos.long:=-pos.long END;   (* west *)

    speed:=0;
    course:=0;
    IF ok THEN
      n:=ORD(buf[payload+4]); 
      IF (n>=28) & (n<=127) THEN speed:=(n-28)*10 ELSE ok:=FALSE END;

      n:=ORD(buf[payload+5]);
      IF (n>=28) & (n<=127) THEN
        DEC(n, 28);
        INC(speed, n DIV 10);
        course:=n MOD 10*100;
      ELSE ok:=FALSE END;

      n:=ORD(buf[payload+6]);
      IF (n>=28) & (n<=127) THEN
        DEC(n, 28);
        INC(course, n);
        IF speed >=800 THEN DEC(speed,  800) END;
        IF course>=400 THEN DEC(course, 400) END;
        IF course>=360 THEN ok:=FALSE END;
      ELSE ok:=FALSE END;

      symb :=buf[payload+7];
      symbt:=Mapsym(buf[payload+8]);
    END;

    compos:=payload+9;

    i:=payload+9;

    IF buf[i+4]="}" THEN                                           (* kenwood ... *)
      IF buf[i]="]" THEN INC(i) END;                               (* TMD700 singularity *)
      IF buf[i]=">" THEN INC(i) END;                               (* THD7 singularity *)
    END;

    IF (buf[i]=CHR(1dH)) OR (buf[i]=CHR(60H)) OR (buf[i]=CHR(27H)) THEN INC(i) END;                   (* hex telemetry bytes *)

    IF buf[i+3]="}" THEN                                          (* altitude in meters +10000 *)
      sc:=0.0;
      IF r91(sc, buf[i],   91.0*91.0) 
       & r91(sc, buf[i+1], 91.0)
       & r91(sc, buf[i+2], 1.0) THEN 
         altitude:=VAL(INTEGER, TRUNC(sc))-10000;
         compos:=i+4;
      ELSE ok:=FALSE END;
    END; 
    postyp:=ENCODEMICE;

  ELSIF c="$" THEN                       (* gps 4806.9409,N,01134.6219,E *)
    ok:=FALSE;
    IF (buf[payload+1]="G") & (buf[payload+2]="P") THEN
      i:=payload+7;
      gpst:=buf[payload+5];
      IF (buf[payload+3]="G") & (buf[payload+4]="G") & (gpst="A") THEN 
        komma(i);
        ok:=TRUE;
      ELSIF (buf[payload+3]="R") & (buf[payload+4]="M") & (gpst="C") THEN 
        komma(i);
        komma(i);
        ok:=TRUE;
      ELSIF (buf[payload+3]="G") & (buf[payload+4]="L") & (gpst="L") THEN 
        ok:=TRUE;
      END;

      pos.lat:=0.0;
               IF NOT dig(pos.lat, buf[i], RAD*10.0) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.lat, buf[i], RAD) THEN ok:=FALSE END;    
      INC(i);  IF NOT dig(pos.lat, buf[i], RAD/6.0) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.lat, buf[i], RAD/60.0) THEN ok:=FALSE END;
      INC(i,2);IF NOT dig(pos.lat, buf[i], RAD/600.0) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.lat, buf[i], RAD/6000.0) THEN ok:=FALSE END;
      IF dig(pos.lat, buf[i+1], RAD/60000.0) THEN INC(i); nornd:=TRUE END;
      IF dig(pos.lat, buf[i+1], RAD/600000.0) THEN INC(i); nornd:=TRUE  END;
      komma(i);
      IF buf[i]="S" THEN pos.lat:=-pos.lat END;
      komma(i);

      pos.long:=0.0;
               IF NOT dig(pos.long, buf[i], RAD*100.0) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.long, buf[i], RAD*10.0) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.long, buf[i], RAD) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.long, buf[i], RAD/6.0) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.long, buf[i], RAD/60.0) THEN ok:=FALSE END;
      INC(i,2);IF NOT dig(pos.long, buf[i], RAD/600.0) THEN ok:=FALSE END;
      INC(i);  IF NOT dig(pos.long, buf[i], RAD/6000.0) THEN ok:=FALSE END;
      IF dig(pos.long, buf[i+1], RAD/60000.0) THEN INC(i); nornd:=TRUE  END;
      IF dig(pos.long, buf[i+1], RAD/600000.0) THEN INC(i); nornd:=TRUE  END;
      komma(i);
      IF buf[i]="W" THEN pos.long:=-pos.long END;

      IF gpst="C" THEN                                                    (* speed and course *)
        komma(i);
        sc:=0.0;
        LOOP
          sc:=sc*10.0;
          IF (i>=len) OR (sc>1000000.0) OR NOT dig(sc, buf[i], 0.1) THEN EXIT END;
          INC(i);
        END;
        speed:=TRUNC(sc/KNOTS+0.5); 

        komma(i);
        sc:=0.0;
        LOOP
          sc:=sc*10.0;
          IF (i>=len) OR (sc>360.0) OR NOT dig(sc, buf[i], 0.1) THEN EXIT END;
          INC(i);
        END;
        course:=TRUNC(sc) MOD 360;

      ELSIF gpst="A" THEN                                                    (* altitude *)
        komma(i);
        komma(i);
        komma(i);
        komma(i);
        sc:=0.0;
        c:=buf[i];
        IF c="-" THEN INC(i) END;
        LOOP
          sc:=sc*10.0;
          IF (i>=len) OR (sc>1000000.0) OR NOT dig(sc, buf[i], 0.1) THEN EXIT END;
          INC(i);
        END;
        komma(i);
        IF c="-" THEN sc:=-sc END;
        IF buf[i]="M" THEN altitude:=VAL(INTEGER, sc) END;
      END;

      WHILE (i<len) & (buf[i]<>"*") DO INC(i) END;
      IF i+3<len THEN compos:=i+3 ELSE compos:=len END; 
      postyp:=ENCODEGPS;
    END;

  ELSE
(* !4915.10N/01402.20E& *)
(* xnnnn.nnNxnnnnn.nnEx *)
    ok:=FALSE;
    pos.lat:=0.0;
    pos.long:=0.0;
    i:=0; 
    IF (buf[payload]="!") OR (buf[payload]="=") THEN i:=payload+1
    ELSIF ((buf[payload]="/") OR (buf[payload]="@"))
    & ((buf[payload+7]="z") OR (buf[payload+7]="h")
    OR (buf[payload+7]="/")) THEN i:=payload+8                          (* skip time *)
    ELSIF buf[payload]=";" THEN i:=payload+18                           (* skip object *)
    ELSIF buf[payload]=")" THEN                                         (* skip item *)
      i:=payload+4; 
      WHILE (i<len) & (buf[i]<>"!") & (buf[i]<>"_") DO INC(i) END;
      INC(i); 
    END;

    IF i>0 THEN
      IF (buf[i]>="0") & (buf[i]<="9") THEN
        IF (i+17<=len)  & (buf[i+4]=".") & (buf[i+14]=".") 
        & ns(buf[i+7]) & ew(buf[i+17]) THEN                            (* not compressed *)
          ok:=TRUE;
                    IF NOT dig(pos.lat, buf[i], 10.0*RAD) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.lat, buf[i], 1.0*RAD) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.lat, buf[i], 10.0*RAD/60.0) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.lat, buf[i], 1.0*RAD/60.0) THEN ok:=FALSE END;
          INC(i,2); IF NOT dig(pos.lat, buf[i], 0.1*RAD/60.0) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.lat, buf[i], 0.01*RAD/60.0) THEN ok:=FALSE END;
          INC(i);   IF CAP(buf[i])="S" THEN pos.lat:=-pos.lat
                    ELSIF CAP(buf[i])<>"N" THEN ok:=FALSE END;
          INC(i);
          symbt:=buf[i];
          INC(i);   IF NOT dig(pos.long, buf[i], 100.0*RAD) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.long, buf[i], 10.0*RAD) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.long, buf[i], 1.0*RAD) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.long, buf[i], 10.0*RAD/60.0) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.long, buf[i], 1.0*RAD/60.0) THEN ok:=FALSE END;
          INC(i,2); IF NOT dig(pos.long, buf[i], 0.1*RAD/60.0) THEN ok:=FALSE END;
          INC(i);   IF NOT dig(pos.long, buf[i], 0.01*RAD/60.0) THEN ok:=FALSE END;
          INC(i);   IF CAP(buf[i])="W" THEN pos.long:=-pos.long
          ELSIF CAP(buf[i])<>"E" THEN ok:=FALSE END;
          INC(i);
          symb:=buf[i];
          INC(i);
          compos:=i;    
       
          IF (i+7<=len) & dig3(nc, i) & (buf[i]="/") THEN            (* ccc/sss *)
            INC(i);
            IF dig3(n, i) THEN
              speed:=n;
              IF (nc>0) & (nc<=360) THEN course:=nc MOD 360 END; 
              compos:=i;
            END;
          END;

          postyp:=ENCODEGPS;
        END;

      ELSE  
        IF i+11+2*ORD(buf[i+10]<>" ")<=len THEN                      (* compressed pos*)
          ok:=TRUE;
          symbt:=Mapsym(buf[i]);
          INC(i); IF NOT r91(pos.lat, buf[i], 91.0*91.0*91.0) THEN ok:=FALSE END;      
          INC(i); IF NOT r91(pos.lat, buf[i], 91.0*91.0) THEN ok:=FALSE END;
          INC(i); IF NOT r91(pos.lat, buf[i], 91.0) THEN ok:=FALSE END;
          INC(i); IF NOT r91(pos.lat, buf[i], 1.0) THEN ok:=FALSE END;
          pos.lat:=90.0*RAD - pos.lat*(RAD/380926.0);
          INC(i); IF NOT r91(pos.long, buf[i], 91.0*91.0*91.0) THEN ok:=FALSE END;
          INC(i); IF NOT r91(pos.long, buf[i], 91.0*91.0) THEN ok:=FALSE END;
          INC(i); IF NOT r91(pos.long, buf[i], 91.0) THEN ok:=FALSE END;
          INC(i); IF NOT r91(pos.long, buf[i], 1.0) THEN ok:=FALSE END;
          pos.long:=pos.long*(RAD/190463.0) - 180.0*RAD;
          symb:=buf[i+1];

          IF ok THEN  
            IF buf[i+2]="}" THEN                           (* radio range *)
            ELSIF ORD(buf[i+4]) DIV 8 MOD 4=2 THEN         (* T byte says GGA so we have altitude *)
              na:=ORD(buf[i+2]);                           (* sc is altitude *)
              n :=ORD(buf[i+3]);
              IF (na>=33) & (na<=127) & (n>=33) & (n<=127) THEN 
                altitude:=TRUNC(FEET*power(1.002, FLOAT((na-33)*91 + n-33)));  (* in feet *)
              END;
            ELSE
              n:=ORD(buf[i+2]);                            (* speed course *)
              IF (n>=33) & (n<=127) THEN
                DEC(n, 33);
                IF n<=89 THEN
                  course:=n*4;
                  n:=ORD(buf[i+3]);
                  IF (n>=33) & (n<=127) THEN speed:=TRUNC(power(1.08, FLOAT(n-33))) END;
                END;
              END;
            END;
            compos:=i+5;
            nornd:=TRUE; 
          END;
          postyp:=ENCODECOMP;
        END;
      END;  
    END;
    
  END;

  IF compos>0 THEN
    i:=compos;                                           (* look for altitude in comment /A= *)
    LOOP
      IF i+9>len THEN EXIT END;

--      n:=i+3;  
--      IF (buf[i]="/") & (buf[i+1]="A") & (buf[i+2]="=") & dig3(na, n) & dig3(nc, n) THEN
      IF (buf[i]="/") & (buf[i+1]="A") & (buf[i+2]="=") & dig6(sc, i+3) THEN

--      sc:=FEET*FLOAT(na*1000 + nc); 
--      IF sc<10000000.0 THEN altitude:=TRUNC(sc) END;

        altitude:=VAL(INTEGER, sc*FEET);
        IF i=compos THEN INC(compos, 9) END;             (* altitude is leftest part of comment *)        
        EXIT
      END;  
      INC(i);
    END;

    IF (pos.long<>0.0) OR (pos.lat<>0.0) THEN 
      i:=compos;                                           (* look for !DAO! precision extension *)
      LOOP
        IF i+5>len THEN EXIT END; 

        IF (buf[i]="!") & (buf[i+4]="!") THEN
          c:=buf[i+1];
          sc:=0.0;
          scl:=0.0;
          IF (c>="A") & (c<="Z") & dig(sc, buf[i+2], DAO10) & dig(scl, buf[i+3], DAO10) THEN 
            IF pos.lat<0.0  THEN pos.lat :=pos.lat -sc  ELSE pos.lat :=pos.lat +sc  END;
            IF pos.long<0.0 THEN pos.long:=pos.long-scl ELSE pos.long:=pos.long+scl END;
            IF i=compos THEN INC(compos, 5) END;
            nornd:=TRUE; 
          ELSIF (c>="a") & (c<="z") & r91(sc, buf[i+2], DAO91) & r91(scl, buf[i+3], DAO91) THEN
            IF pos.lat<0.0  THEN pos.lat :=pos.lat-sc   ELSE pos.lat:=pos.lat+sc    END;
            IF pos.long<0.0 THEN pos.long:=pos.long-scl ELSE pos.long:=pos.long+scl END;
            IF i=compos THEN INC(compos, 5) END;
            nornd:=TRUE;
          END;
          IF postyp>0C THEN postyp:=CAP(postyp) END;
          EXIT

        END;
        INC(i);
      END;
    END;
  END;
  
  IF ok & NOT nornd THEN                                               (* rounding *)
    IF pos.lat<0.0 THEN pos.lat:=pos.lat-RND
    ELSIF pos.lat>0.0 THEN pos.lat:=pos.lat+RND END;
    IF pos.long<0.0 THEN pos.long:=pos.long-RND
    ELSIF pos.long>0.0 THEN pos.long:=pos.long+RND END;
  END;


  ok:=ok & ((pos.long<>0.0) OR (pos.lat<>0.0)) 
         & (ABS(pos.long)<180.0*RAD) & (ABS(pos.lat)<90.0*RAD);
  IF NOT ok THEN pos.long:=0.0; pos.lat:=0.0 END;

END GetPos;


PROCEDURE GetSym(d-:ARRAY OF CHAR; VAR symb, symt:CHAR);   (* symbol out of destination call *)
CONST SYMT1="BBBCBDBEBFBGBHBIBJBKBLBMBNBOBPP0P1P2P3P4P5P6P7P8P9MRMSMTMUMVMWMXPAPBPCPDPEPFPGPHPIPJPKPLPMPNPOPPPQPRPSPTPUPVPWPXPYPZHSHTHUHVHWHXLALBLCLDLELFLGLHLILJLKLLLMLNLOLPLQLRLSLTLULVLWLXLYLZJ1J2J3J4";
      SYMT2="OBOCODOEOFOGOHOIOJOKOLOMONOOOPA0A1A2A3A4A5A6A7A8A9NRNSNTNUNVNWNXAAABACADAEAFAGAHAIAJAKALAMANAOAPAQARASATAUAVAWAXAYAZDSDTDUDVDWDXSASBSCSDSESFSGSHSISJSKSLSMSNSOSPSQSRSSSTSUSVSWSXSYSZQ1Q2Q3Q4";
      LEN=94;
 
VAR n:CARDINAL;

BEGIN
  IF (d[0]="G") & (d[1]="P") & (d[2]="S") & ((d[3]="C") OR (d[3]="E")) THEN
    IF (d[4]>="0") & (d[4]<="9") & (d[5]>="0") & (d[5]<="9") THEN
      n:=(ORD(d[4])-ORD("0"))*10 + ORD(d[5])-ORD("0");
      IF (n>=1) & (n<=94) THEN 
        symb:=CHR(32+n);
        IF d[3]="C" THEN symt:="/" ELSE symt:="\" END;
      END;
    END;
  ELSIF (d[0]="G") & (d[1]="P") & (d[2]="S")
  OR (d[0]="S") & (d[1]="P") & (d[2]="C")
  OR (d[0]="S") & (d[1]="Y") & (d[2]="M") THEN
    IF d[3]=0C THEN symb:="/"; symt:="/";
    ELSE
      n:=0;
      LOOP
        IF n>=LEN THEN EXIT END;

        IF (d[3]=SYMT1[n*2]) & (d[4]=SYMT1[n*2+1]) THEN
          symb:=CHR(33 + n);
          symt:="/";
          EXIT
        END;

        IF (d[3]=SYMT2[n*2]) & (d[4]=SYMT2[n*2+1]) THEN
          symb:=CHR(33 + n);
          symt:="\";
          IF (d[5]>="0") & (d[5]<="9") OR (d[5]>="A") & (d[5]<="Z") THEN symt:=d[5] END; (* overlay *)
          EXIT
        END;

        INC(n);
      END;
    END;
  END;
END GetSym;



END aprspos.
